// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"

	"github.com/m4rw3r/uuid"
	"github.com/spf13/cobra"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// actorCmd represents the actor command
var actorCmd = &cobra.Command{
	Use:   "actor",
	Short: "A brief description of your command",
	Long: `A longer description that spans multiple lines and likely contains examples
and usage of using your command. For example:

Cobra is a CLI library for Go that empowers applications.
This application is a tool to generate the needed files
to quickly create a Cobra application.`,
	//Run: func(cmd *cobra.Command, args []string) {
	// TODO: Work your own magic here
	//fmt.Println("actor called")
	//},
}

// listCmd represents the list command
var actorCreateCmd = &cobra.Command{
	Use:   "create",
	Short: "Create a new actor",
	Long:  ``,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient()
		if err != nil {
			return
		}
		id, _ := uuid.FromString(cmd.Flag("id").Value.String())
		actor := api.Actor{
			ID:    id,
			Name:  cmd.Flag("name").Value.String(),
			Roles: []string{cmd.Flag("type").Value.String()},
		}
		err = actor.Kind.UnmarshalText([]byte(cmd.Flag("type").Value.String()))
		if err != nil {
			Logger.Errorf("Error parsing type. Got %s", err)
			return
		}
		created, err := c.ActorCreate([]api.Actor{actor})
		if err != nil {
			Logger.Errorf("Error: %s", err)
			return
		}
		if len(created) == 0 {
			fmt.Println("No actor created")
		} else {
			for _, actor := range created {
				fmt.Println(fmtActor("", actor))
			}
		}
	},
}

// listCmd represents the list command
var actorListCmd = &cobra.Command{
	Use:   "list",
	Short: "",
	Long:  ``,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient()
		if err != nil {
			return
		}
		id, _ := uuid.FromString(cmd.Flag("id").Value.String())
		actor := api.Actor{
			ID:    id,
			Name:  cmd.Flag("name").Value.String(),
			Roles: []string{cmd.Flag("role").Value.String()},
		}
		if cmd.Flag("type").Value.String() != "" {
			err = actor.Kind.UnmarshalText([]byte(cmd.Flag("type").Value.String()))
			if err != nil {
				Logger.Errorf("Error parsing type. Got %s", err)
				return
			}
		}
		actorList := []api.Actor{}
		if !actor.ID.IsZero() || actor.Name != "" || !actor.AccountID.IsZero() {
			actorList = append(actorList, actor)
		}
		actors, err := c.ActorList(actorList)
		if err != nil {
			Logger.Errorf("Error: %s", err)
			return
		}
		if len(actors) == 0 {
			fmt.Println("No actors found")
		} else {
			for _, actor := range actors {
				fmt.Println(fmtActor("", actor))
			}
		}
	},
}

// actorDeleteCmd represents the actor delete sub-command
var actorDeleteCmd = &cobra.Command{
	Use:   "delete",
	Short: "Delete an actor",
	Long: "Delete an actor from the xbus server. You must restart " +
		"the xbus clients for this change to be taken into account.",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			Logger.Errorf("expects one and only one identifier")
			cmd.Usage()
			os.Exit(-1)
		}
		os.Exit(doDeleteActor(args[0]))
	},
}

func doDeleteActor(identifier string) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}
	actor, err := conn.ActorFind(identifier)
	if err != nil {
		if err == xbus.ErrNotFound {
			Logger.Errorf("actor not found")
		} else {
			Logger.Errorf("could not find actor: %s", err)
		}
		return -1
	}
	if err := conn.ActorDelete([]uuid.UUID{actor.ID}); err != nil {
		Logger.Errorf("could not delete actor: %s", err)
		return -1
	}
	return 0
}

// actorAcceptCmd represents the actor accept sub-command.
var actorAcceptCmd = &cobra.Command{
	Use:   "accept",
	Short: "Accept an actor.",
	Long: "Accept a pending actor so it can be allowed to connect to Xbus. An actor ID or name " +
		"is required.",
	Run: acceptActor,
}

// acceptActor parses arguments to accept a pending actor.
func acceptActor(cmd *cobra.Command, args []string) {
	var identifier string
	if v := cmd.Flag("id"); v.Changed {
		identifier = v.Value.String()
	} else if len(args) == 1 {
		identifier = args[0]
	} else {
		cmd.Usage()
		os.Exit(1)
	}
	os.Exit(doAcceptActor(identifier))
}

// doAcceptActor accepts a pending actor request.
func doAcceptActor(identifier string) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	actor, err := conn.ActorFind(identifier)
	if err != nil {
		if err == xbus.ErrNotFound {
			Logger.Errorf("actor not found")
		} else {
			Logger.Errorf("could not find actor: %s", err)
		}
		return -1
	}

	// TODO Find the actor based on the name.

	// Send an "accept" request.
	repActor, err := conn.ActorAccept(actor.ID)
	if err != nil {
		Logger.Errorf("Could not accept the actor %q: %s", actor.ID, err)
		return 1
	}
	if repActor.Status == api.ActorPending {
		Logger.Errorf("Could not accept the actor %q: Still pending.", actor.ID)
		return 1
	}
	Logger.Noticef("Successfully accepted the actor %q (ID %q).", repActor.Name, actor.ID)
	return 0
}

// actorRejectCmd represents the actor reject sub-command.
var actorRejectCmd = &cobra.Command{
	Use:   "reject",
	Short: "Reject an actor.",
	Long: "Reject a pending actor so it can be allowed to connect to Xbus. An actor ID or name " +
		"is required.",
	Run: rejectActor,
}

// rejectActor parses arguments to reject a pending actor.
func rejectActor(cmd *cobra.Command, args []string) {
	var identifier string
	if v := cmd.Flag("id"); v.Changed {
		identifier = v.Value.String()
	} else if len(args) == 1 {
		identifier = args[0]
	} else {
		cmd.Usage()
		os.Exit(1)
	}
	os.Exit(doRejectActor(identifier))
}

// doRejectActor rejects a pending actor request.
func doRejectActor(identifier string) int {
	if len(identifier) == 0 {
		Logger.Errorf("Please provide an actor identifier.")
		return 1
	}

	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	actor, err := conn.ActorFind(identifier)
	if err != nil {
		if err == xbus.ErrNotFound {
			Logger.Errorf("actor not found")
		} else {
			Logger.Errorf("could not find actor: %s", err)
		}
		return -1
	}

	// Send an "reject" request.
	repActor, err := conn.ActorReject(actor.ID)
	if err != nil {
		Logger.Errorf("Could not reject the actor %q: %s", identifier, err)
		return 1
	}
	if repActor.Status == api.ActorPending {
		Logger.Errorf("Could not reject the actor %q: Still pending.", identifier)
		return 1
	}
	Logger.Noticef("Successfully rejected the actor %q (ID %q).", repActor.Name, identifier)
	return 0
}

func fmtActor(prefix string, actor api.Actor) string {
	s := fmt.Sprintf(
		"%sID: %s\n%sName: %s\n%sKind: %s\n%sRole: %s\n",
		prefix, actor.ID,
		prefix, actor.Name,
		prefix, actor.Kind,
		prefix, actor.Roles[0])
	return s
}

func init() {
	RootCmd.AddCommand(actorCmd)
	actorCmd.AddCommand(actorCreateCmd)
	actorCmd.AddCommand(actorListCmd)
	actorCmd.AddCommand(actorAcceptCmd)
	actorCmd.AddCommand(actorRejectCmd)
	actorCmd.AddCommand(actorDeleteCmd)

	flags := actorCmd.PersistentFlags()
	flags.String("id", "", "Actor's ID")
	flags.String("name", "", "Actor's name")
	flags.String("role", "", "Actor's role")
	flags.String("type", "", "Actor's type (emitter, consumer)")
}
