package process

import (
	"fmt"
	"time"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/storage"
)

// StateEngine provides api to consistently modify a process state
type StateEngine struct {
	process *Process
	state   storage.ProcessState
	graph   graph.Graph
	log     xbus.Logger
}

// GetState returns the current state (which is expected to be consistent)
func (e *StateEngine) GetState() storage.ProcessState {
	return e.state
}

// GetNodeState returns a copy of a node state and the node graph definition
func (e *StateEngine) GetNodeState(
	node storage.ProcessNode,
) (storage.ProcessNodeState, storage.Node) {
	graphNode, err := e.graph.GetNode(node.NodeID)
	if err != nil {
		panic(err)
	}
	nodeState := e.state.NodeStates[node]
	if nodeState.InputStatuses == nil {
		nodeState.InputStatuses = make(storage.ProcessNodeIOStatusMap)
		if len(graphNode.Inputs) == 0 {
			nodeState.Input = storage.ProcessNodeIODone
		}
	}
	if nodeState.OutputStatuses == nil {
		nodeState.OutputStatuses = make(storage.ProcessNodeIOStatusMap)
	}
	return nodeState, graphNode
}

// SetNodeState updates a node state
func (e *StateEngine) SetNodeState(node storage.ProcessNode, state storage.ProcessNodeState) {
	e.state.NodeStates[node] = state

	e.updateProcessStatus()
}

// updateNodeStatus determinate the next node status based on its current
// input and output states.
func (e *StateEngine) updateNodeStatus(nodeState *storage.ProcessNodeState) {
	switch nodeState.Status {
	case storage.ProcessNodeDone, storage.ProcessNodeUnreachable, storage.ProcessNodeError:
		// The status cannot change anymore, no matter what the inputs/outputs
		// are.
		return
	}
	nodeState.Status = ioStatusMatrix[nodeState.Output][nodeState.Input]
}

func (e *StateEngine) updateInputStatus(
	graphNode storage.Node,
	nodeState *storage.ProcessNodeState,
	input string,
	status storage.ProcessNodeIOStatus,
) {
	nodeState.InputStatuses[input] = status
	nodeState.Input = combineIOStatus(graphNode.Inputs, nodeState.InputStatuses)

	e.updateNodeStatus(nodeState)
}

func (e *StateEngine) updateOuputStatus(
	graphNode storage.Node,
	nodeState *storage.ProcessNodeState,
	output string,
	status storage.ProcessNodeIOStatus,
) {
	nodeState.OutputStatuses[output] = status
	nodeState.Output = combineIOStatus(graphNode.Outputs, nodeState.OutputStatuses)

	e.updateNodeStatus(nodeState)
}

func (e *StateEngine) updateProcessStatus() {
	anyError := false
	hitNodes := make(map[string]bool)
	for pNode, state := range e.state.NodeStates {
		hitNodes[pNode.NodeID] = true
		if state.Status == storage.ProcessNodeError {
			anyError = true
		}
	}

	if anyError {
		e.state.Status = api.ProcessError
		return
	}

	allHit := true
	for _, node := range e.graph.Nodes {
		hit, _ := hitNodes[node.ID]
		if !hit {
			allHit = false
			break
		}
	}

	if e.state.Status == api.ProcessInitial {
		e.state.Status = api.ProcessRunning
	}
	if allHit {
		allDone := true
		for _, nState := range e.state.NodeStates {
			if nState.Status != storage.ProcessNodeDone &&
				nState.Status != storage.ProcessNodeUnreachable {
				allDone = false
			}
		}
		if allDone {
			e.state.Status = api.ProcessDone
		}
	}
}

// HitInput signal a data emission to an input
func (e *StateEngine) HitInput(
	node storage.ProcessNode,
	input string,
	complete bool,
	closed bool,
	timestamp time.Time,
) {
	initialStatus := e.state.Status
	nodeState, graphNode := e.GetNodeState(node)

	iostatus := nodeState.InputStatuses[input]

	if iostatus == storage.ProcessNodeIODone {
		iostatus = storage.ProcessNodeIOError
	} else if iostatus != storage.ProcessNodeIOError {
		if complete {
			if closed {
				iostatus = storage.ProcessNodeIOClosed
			} else {
				iostatus = storage.ProcessNodeIODone
			}
		} else {
			iostatus = storage.ProcessNodeIOStarted
		}
	}

	e.updateInputStatus(graphNode, &nodeState, input, iostatus)

	if iostatus == storage.ProcessNodeIOClosed && len(nodeState.InputStatuses) > 1 {
		panic("Case not handled: Closing input of a multiple inputs processnode")
	}

	e.SetNodeState(node, nodeState)

	// Single input was closed: closing all the outputs
	if iostatus == storage.ProcessNodeIOClosed {
		if len(graphNode.Outputs) == 0 {
			// no outputs, force output to 'done' so the node state
			// will be automaticaly marked done.
			nodeState.Output = storage.ProcessNodeIODone
			e.updateNodeStatus(&nodeState)
			e.SetNodeState(node, nodeState)
		} else {
			for _, output := range graphNode.Outputs {
				e.HitOutput(node, output, storage.ReceptionClosed, timestamp)
			}
		}
	}

	e.state.NodeStatesChanged = timestamp

	if initialStatus != e.state.Status {
		e.state.StatusChanged = timestamp
	}
}

// HitOutput signal a data reception from an output
func (e *StateEngine) HitOutput(
	node storage.ProcessNode,
	output string,
	status storage.ReceptionStatus,
	timestamp time.Time,
) {
	initialStatus := e.state.Status

	nodeState, graphNode := e.GetNodeState(node)

	iostatus := nodeState.OutputStatuses[output]

	if iostatus == storage.ProcessNodeIODone {
		iostatus = storage.ProcessNodeIOError
	} else if iostatus != storage.ProcessNodeIOError {
		if status == storage.ReceptionDone {
			iostatus = storage.ProcessNodeIODone
		} else if status == storage.ReceptionClosed {
			iostatus = storage.ProcessNodeIOClosed
		} else {
			iostatus = storage.ProcessNodeIOStarted
		}
	}

	e.updateOuputStatus(graphNode, &nodeState, output, iostatus)

	e.SetNodeState(node, nodeState)

	// If the output was closed, propagate to the next nodes in line
	// TODO here we directly mark the next inputs as closed, which means the actual
	// actors will never know about it.
	// But in the case of multiple inputs, the actor will need to know if some inputs
	// were closed.
	// We also need to decide if some actors may be notified that all their inputs
	// were closed in a particular process. It may make sens someday.
	if iostatus == storage.ProcessNodeIOClosed {
		targets, err := e.process.GetTargets(storage.NodeOutputRef{
			NodeID: graphNode.ID,
			Output: output,
		}, uuid.UUID{}, true)
		if err != nil {
			panic(err)
		}
		for _, target := range targets {
			e.HitInput(
				storage.ProcessNode{
					NodeID:  target.NodeID,
					ActorID: target.ActorID,
				},
				target.Input,
				true, true,
				timestamp)
		}
	}

	e.state.NodeStatesChanged = timestamp

	if initialStatus != e.state.Status {
		e.state.StatusChanged = timestamp
	}
}

// EndProcessNode signal the end of processing by a process node
func (e *StateEngine) EndProcessNode(
	node storage.ProcessNode,
	result api.ActorProcessingResultType,
	timestamp time.Time,
) {
	nodeState, graphNode := e.GetNodeState(node)

	// Update the node status
	if nodeState.Status != storage.ProcessNodeActive {
		panic("Cannot end a processnode that is not in 'active' state. Current state: " + nodeState.Status.String())
	}
	switch result {
	case api.ActorProcessingSuccess:
		nodeState.Status = storage.ProcessNodeDone
	case api.ActorProcessingError:
		nodeState.Status = storage.ProcessNodeError
	default:
		panic(fmt.Sprintf("Invalid ActorProcessingResultType: %s", result))
	}
	e.SetNodeState(node, nodeState)

	// Close remaining outputs
	// XXX what about inputs ? Currently they will not stop from force ending a
	// node, but they will prevent it from being auto ended by updateProcessStatus
	for _, output := range graphNode.Outputs {
		switch nodeState.OutputStatuses[output] {
		case storage.ProcessNodeIOPending:
			e.HitOutput(node, output, storage.ReceptionClosed, timestamp)
		case storage.ProcessNodeIOStarted:
			// TODO report the error somewhere, somehow
			e.HitOutput(node, output, storage.ReceptionError, timestamp)
		default:
			// Ouput already closed, do nothing
		}
	}
}

func combineIOStatus(ionames []string, m storage.ProcessNodeIOStatusMap) storage.ProcessNodeIOStatus {
	var (
		started    = false
		allStarted = true
		allClosed  = true
		done       = true
		anyerror   = false
	)
	for _, name := range ionames {
		switch m[name] {
		case storage.ProcessNodeIOPending:
			allStarted = false
			allClosed = false
			done = false
		case storage.ProcessNodeIOStarted:
			started = true
			allClosed = false
			done = false
		case storage.ProcessNodeIOClosed:
			started = true
		case storage.ProcessNodeIODone:
			allClosed = false
			started = true
		case storage.ProcessNodeIOError:
			allClosed = false
			allStarted = false
			done = false
			anyerror = true
		default:
			panic("Invalid io status for single io")
		}
	}
	status := storage.ProcessNodeIOPending
	if anyerror {
		status = storage.ProcessNodeIOError
	} else if done {
		status = storage.ProcessNodeIODone
	} else if allStarted {
		status = storage.ProcessNodeIOStarted
	} else if allClosed {
		status = storage.ProcessNodeIOClosed
	} else if started {
		status = storage.ProcessNodeIOAllStarted
	}
	return status
}
