// Copyright 2016 Orus.io. All rights reserved.

package control

import (
	"gopkg.in/yaml.v2"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/graph"
)

func (s *Service) handleGraphQuery(subject, reply string, req api.ControlGraphQueryReq) {
	var rep api.ControlGraphQueryRep

	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	graphs := s.o.GraphManager.Query(req.Name, "", false)
	for _, graph := range graphs {
		rep.GraphList = append(rep.GraphList,
			api.GraphInfo{
				ID:      graph.ID,
				Name:    graph.Name,
				Version: graph.Version,
				Status:  graph.Status,
			})
	}
	rep.Status = api.StatusOK
}

func (s *Service) handleGraphSave(subject, reply string, req api.ControlGraphSaveReq) {
	var rep api.ControlGraphSaveRep

	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	var graph graph.Graph
	if err := yaml.Unmarshal([]byte(req.Graph), &graph); err != nil {
		rep.Response = api.ClientError("Error decoding graph: %s", err)
		return
	}

	if req.Info.ID.IsZero() {
		graphs := s.o.GraphManager.Query(req.Info.Name, req.Info.Version, false)
		if len(graphs) > 1 {
			rep.Response = api.ClientError(
				"Multiple graph matches: %s-%s", req.Info.Name, req.Info.Version)
			return
		}
		if len(graphs) == 1 {
			graph.ID = graphs[0].ID
		}
		graph.Name = req.Info.Name
		graph.Version = req.Info.Version
	} else {
		curGraph, exists := s.o.GraphManager.GetGraph(req.Info.ID)
		if exists {
			rep.Response = api.ClientError(
				"No such graph: %s", req.Info.ID)
			return
		}
		graph.ID = req.Info.ID
		graph.Name = curGraph.Name
		graph.Version = curGraph.Version
	}

	for _, e := range graph.Validate() {
		rep.ValidationMessages = append(rep.ValidationMessages, e.Error())
	}

	for _, e := range graph.ValidateActors(s.o.ActorStorage) {
		rep.ValidationMessages = append(rep.ValidationMessages, e.Error())
	}

	graph, err := s.o.GraphManager.Save(graph)
	if err != nil {
		rep.Response = api.ClientError("Save graph error: %s", err)
		return
	}

	b, err := yaml.Marshal(map[string]interface{}{
		"nodes": graph.Nodes,
		"edges": graph.Edges,
	})
	if err != nil {
		rep.Response = api.ServerError(err)
		return
	}
	rep.Status = api.StatusOK
	rep.Graph = string(b)
	rep.Info = api.GraphInfo{
		ID:      graph.ID,
		Name:    graph.Name,
		Version: graph.Version,
		Status:  graph.Status,
	}
}

func (s *Service) handleGraphSetStatus(subject, reply string, req api.ControlGraphSetStatusReq) {
	var rep = api.ControlGraphSetStatusRep{
		Response: api.Response{Status: api.StatusOK}}

	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	var info = api.GraphInfo{ID: req.ID}
	var g graph.Graph
	if info.ID.IsZero() {
		var graphs = s.o.GraphManager.Query(req.Name, req.Version, false)
		if len(graphs) == 0 {
			rep.Response = api.ClientError(
				"No such graph: %s-%s", req.Name, req.Version)
			return
		}
		if len(graphs) > 1 {
			rep.Response = api.ClientError("Multiple graphs found")
		}
		g = graphs[0]
		info.ID = g.ID
		info.Name = g.Name
		info.Version = g.Version
		info.Status = g.Status
	} else {
		var ok bool
		g, ok = s.o.GraphManager.GetGraph(info.ID)
		if !ok {
			rep.Response = api.ClientError(
				"No such graph: %s", info.ID)
		}
		info.ID = g.ID
		info.Name = g.Name
		info.Version = g.Version
		info.Status = g.Status
	}
	rep.Info = info
	if err := s.o.GraphManager.SetStatus(g, req.Status); err != nil {
		if list, ok := err.(graph.ValidationErrorList); ok {
			rep.Response = api.ClientError("invalidgraph", err)
			for _, e := range list {
				rep.ValidationMessages = append(rep.ValidationMessages, e.Error())
			}
		} else {
			rep.Response = api.ClientError("%s", err)
		}
	} else {
		rep.Info.Status = req.Status
	}
}
