// Copyright 2016 Orus.io. All rights reserved.

package emission

import (
	"fmt"
	"runtime/debug"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// EmissionQueue is the name we use to queueSub to a queue with the emission
	EmissionQueue = "emission_queue"
)

// Options is used to create a Emission
type Options struct {
	services.Options

	ReceptionTimeout time.Duration
	// CleanupInterval is the interval between each cleanup
	CleanupInterval time.Duration
	// CleanupDelay is how long a emission should be stalled to get cleaned up
	CleanupDelay time.Duration

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	ProcessManager  *process.Manager
}

// NewService creates a Emission
func (o Options) NewService() *Service {
	if o.ReceptionTimeout == 0 {
		o.ReceptionTimeout = 2 * time.Second
	}
	if o.CleanupInterval == 0 {
		o.CleanupInterval = time.Minute
	}
	if o.CleanupDelay == 0 {
		o.CleanupDelay = 10 * time.Minute
	}

	return &Service{
		BaseService: services.NewBaseService("emission", o.Options),
		o:           o,
		quit:        make(chan struct{}),
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	o Options

	quit chan struct{}
}

// Startup starts the service
func (s *Service) Startup() error {
	s.Log.Noticef("Startup")

	if err := s.SubscribeAll(
		services.QueueSubscribe(api.CoreActorEventsSubject, EmissionQueue, s.handleActorEvent),
		services.QueueSubscribe(api.CoreProcessEventsSubject, EmissionQueue, s.handleProcessEvent),
	); err != nil {
		s.Shutdown()
		return err
	}

	signals := make(chan api.EmissionSignal)
	sub, err := s.Conn.BindRecvQueueChan(api.EmissionSignalSubjet, EmissionQueue, signals)
	if err != nil {
		s.Log.Errorf("Cannot subscribe: %s", err)
		return err
	}
	s.AddSubscription(sub)

	go s.handleSignalWorker(signals)

	// if e.isAlone(), reschedule pending emissions

	return nil
}

// Shutdown signal the routines to stop and unsubscribe
func (s *Service) Shutdown() error {
	err := s.UnsubscribeAll()
	close(s.quit)
	s.Tasks.Wait()
	return err
}

func (s *Service) handleActorEvent(subject string, event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "up" {
		// if "up", reschedule emissions
		s.Log.Noticef("%s is back online", event.Actor.ID)
		emissions := s.o.ProcessManager.ListActiveEmissions(event.Actor.ID)

		for _, pending := range emissions {
			s.Log.Noticef("Rescheduling %s to %s", pending.EnvelopeID, pending.TargetActor)
			s.Conn.Publish(api.EmissionSignalSubjet, api.EmissionSignal{
				ProcessID:  pending.ProcessID,
				EnvelopeID: pending.EnvelopeID,
				Target: api.EnvelopeTarget{
					ActorID: pending.TargetActor,
					NodeID:  pending.TargetInputRef.NodeID,
					Input:   pending.TargetInputRef.Input,
				},
			})
		}
	}
}

func (s *Service) handleProcessEvent(subject string, event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "status" &&
		event.Process.OldStatus == api.ProcessPaused &&
		event.Process.Status == api.ProcessRunning {
		s.Log.Noticef("Process %s just restarted", event.Process.ID)

		emissions := s.o.ProcessManager.ListProcessEmissions(event.Process.ID)

		for _, pending := range emissions {
			s.Log.Noticef("Rescheduling %s to %s", pending.EnvelopeID, pending.TargetActor)
			s.Conn.Publish(api.EmissionSignalSubjet, api.EmissionSignal{
				ProcessID:  pending.ProcessID,
				EnvelopeID: pending.EnvelopeID,
				Target: api.EnvelopeTarget{
					ActorID: pending.TargetActor,
					NodeID:  pending.TargetInputRef.NodeID,
					Input:   pending.TargetInputRef.Input,
				},
			})
		}
	}
}

type routineActivity struct {
	signal api.EmissionSignal
	time   time.Time
}

func (s *Service) handleSignalWorker(signals <-chan api.EmissionSignal) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	type emitRoutine struct {
		c            chan bool
		lastActivity time.Time
	}
	emitRoutines := make(map[api.EmissionSignal]emitRoutine)
	// The big chan size is because we never want writting to this channel
	// to be blocking
	// We make the assumption that no more that 1024 activity signals
	// can be acculated before being handled
	routineActivity := make(chan routineActivity, 1024)

	timer := time.NewTimer(s.o.CleanupInterval)
	defer timer.Stop()

	for {
		select {
		case activity := <-routineActivity:
			routine, ok := emitRoutines[activity.signal]
			if ok {
				if (activity.time == time.Time{}) {
					close(routine.c)
					delete(emitRoutines, activity.signal)
				} else {
					routine.lastActivity = activity.time
					emitRoutines[activity.signal] = routine
				}
			}
		case signal := <-signals:
			routine, ok := emitRoutines[signal]
			if ok {
				routine.c <- true
			} else {

				routine.c = make(chan bool)
				emitRoutines[signal] = routine

				process, err := s.o.ProcessManager.Get(signal.ProcessID)
				if err != nil {
					s.Log.Errorf("Received a wrong processID: %s", err)
					continue
				}

				s.TaskAdd()
				go s.emit(routine.c, routineActivity, process, signal.EnvelopeID, signal.Target)
			}
		case t := <-timer.C:
			liveRoutines := make(map[api.EmissionSignal]emitRoutine)
			t = t.Add(-s.o.CleanupDelay)
			for s, r := range emitRoutines {
				if r.lastActivity.Before(t) {
					close(r.c)
				} else {
					liveRoutines[s] = r
				}
			}
			emitRoutines = liveRoutines
		case <-s.quit:
			for _, routine := range emitRoutines {
				close(routine.c)
			}
			return
		}
	}
}

// emit a fragment of envelope to the target.
func (s *Service) emitFragment(process *process.Process, envelopeID uuid.UUID, target api.EnvelopeTarget, actor api.Actor) (sent bool, complete bool, err error) {
	status := process.GetStatus()
	if status != api.ProcessRunning {
		s.Log.Noticef("Process not running (current state: %s), skip emission", status)
		return false, true, nil
	}

	emitPosition, err := process.GetEmitPosition(envelopeID, target)
	if err != nil {
		return false, false, fmt.Errorf("Cannot get emit position: %s", err)
	}
	defer emitPosition.Unlock()

	pos, err := emitPosition.Get()
	if err != nil {
		return false, false, fmt.Errorf("Cannot read emit position: %s", err)
	}

	if pos.Complete {
		s.Log.Debugf("Emission already complete, skipping")
		return false, true, nil
	}

	newenv, newpos, err := s.o.EnvelopeStorage.ReadEnvelope(envelopeID, pos, api.EnvelopeMaxSize)
	if err != nil {
		return false, false, fmt.Errorf("Error reading envelope for transfer: %s", err)
	}
	if newpos.Equals(pos) {
		s.Log.Debugf("No new data to emit")
		return false, false, nil
	}

	newenv.Context = api.EnvelopeContext{
		ProcessID: process.ID,
		NodeID:    target.NodeID,
	}

	var ack api.EnvelopeAck
	outsubject := api.ActorInputSubject(target.ActorID, target.Input)
	// For the non-first fragments of envelope, publish to a dedicated subject
	if !pos.Start {
		outsubject += "." + envelopeID.String()
	}
	if actor.AccountID.IsZero() {
		err = s.Conn.Request(outsubject, newenv, &ack, s.o.ReceptionTimeout)
	} else {
		err = s.Conn.RequestFor(actor.AccountID.String(), outsubject, newenv, &ack, s.o.ReceptionTimeout)
	}
	if err == nats.ErrTimeout {
		err := fmt.Errorf(
			"Actor '%s' did not ack the envelope within %d seconds on %s. "+
				"Mark it unresponsive.",
			actor.Name, s.o.ReceptionTimeout/time.Second, outsubject)
		if err := s.o.ActorStorage.MarkUnresponsive(target.ActorID); err != nil {
			panic(fmt.Sprintf("Cannot mark the actor unresponsive: %s", err))
		}
		return false, false, err
	} else if err != nil {
		return false, false, fmt.Errorf(
			"Error transmitting envelope to %s on %s: %s",
			target, outsubject, err)
	}
	if ack.ReceptionStatus == api.ReceptionError {
		return false, false, fmt.Errorf("Actor returned an error: %s", ack.Reason)
	}
	s.Log.Noticef("Published to %s", outsubject)
	if err := emitPosition.Set(newpos); err != nil {
		return false, false, fmt.Errorf("Cannot set emit position: %s", err)
	}
	return true, pos.Complete, nil
}

func (s *Service) emit(c <-chan bool, activity chan<- routineActivity,
	process *process.Process, envelopeID uuid.UUID, target api.EnvelopeTarget,
) {
	emissionSignal := api.EmissionSignal{
		ProcessID:  process.ID,
		EnvelopeID: envelopeID,
		Target:     target,
	}
	defer s.SafeTaskDone()

	defer func() {
		if r := recover(); r != nil {
			s.Log.Errorf("Error in emit: %s\n%s", r, debug.Stack())
		}
	}()

	actor, ok := s.o.ActorStorage.Get(target.ActorID)
	if !ok {
		s.Log.Errorf("Unknown actor")
		return
	}

	for {
		// Emit as much of the envelope we can
		for {
			sent, complete, err := s.emitFragment(process, envelopeID, target, actor)
			if err != nil {
				s.Log.Errorf("%s", err)
			}
			if err != nil || complete {
				activity <- routineActivity{emissionSignal, time.Time{}}
			} else {
				activity <- routineActivity{emissionSignal, time.Now()}
			}
			if !sent {
				break
			}
		}

		// wait for a signal
		select {
		case continuing := <-c:
			if !continuing {
				return
			}
		}
	}
}
