// Copyright 2016 Orus.io. All rights reserved.

package mocks

import (
	"fmt"
	"math"
	"sync"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewEnvelopeStorage init an empty EnvelopeStorage suitable for tests
func NewEnvelopeStorage(logger xbus.Logger) *EnvelopeStorage {
	return &EnvelopeStorage{
		receivers: make(map[uuid.UUID]*xbus.EnvelopeReceiver),
		envelopes: make(map[uuid.UUID]api.Envelope),
		logger:    logger,
	}
}

// EnvelopeStorage implements EnvelopeStorage on top of a SQL database
type EnvelopeStorage struct {
	receivers map[uuid.UUID]*xbus.EnvelopeReceiver
	envelopes map[uuid.UUID]api.Envelope
	logger    xbus.Logger

	m sync.Mutex
}

// GetEventTypes returns an envelop event types
func (s *EnvelopeStorage) GetEventTypes(id uuid.UUID) ([]string, bool) {
	r, ok := s.receivers[id]
	if !ok {
		return nil, false
	}
	eventtypes := r.EventTypes()
	return eventtypes, len(eventtypes) != 0
}

func (s *EnvelopeStorage) receiveEnvelope(id uuid.UUID) {
	envelope, ok := s.envelopes[id]
	r := s.receivers[id]

	if !ok {
		envelope.ID = r.ID
	}

	eventtypes := r.EventTypes()
	if eventtypes == nil {
		return
	}

	if len(envelope.Events) == 0 {
		envelope.Events = make([]api.Event, len(eventtypes))
	}

	for i, event := range r.Events() {
		envelope.Events[i].ID = event.ID
		envelope.Events[i].Type = event.Type
		envelope.Events[i].ItemCount = event.ItemCount
		envelope.Events[i].Checksum = event.Checksum
		envelope.Events[i].Items = append(
			envelope.Events[i].Items,
			r.NextItems(event.Type, int(math.MaxInt64))...)
	}

	s.envelopes[id] = envelope
}

// StoreEnvelope store a part of an envelope and returns its status
// in the storage
func (s *EnvelopeStorage) StoreEnvelope(e api.Envelope) (storage.EnvelopeStatus, error) {
	s.m.Lock()
	defer s.m.Unlock()

	var status = storage.EnvelopeStatus{ID: e.ID}

	r, ok := s.receivers[e.ID]
	if !ok {
		newr, err := xbus.NewEnvelopeReceiver("", e)
		if err != nil {
			status.Status = "error"
			status.Err = err
			return status, err
		}
		s.receivers[e.ID] = newr
		r = newr
	} else {
		if err := r.AddFragment(e); err != nil {
			status.Status = "error"
			status.Err = err
			return status, err
		}
	}

	status.TypesKnown = r.EventTypes() != nil
	switch r.Status() {
	case api.ReceptionReceiving:
		status.Status = "receiving"
	case api.ReceptionAccepted:
		status.Status = "complete"
	case api.ReceptionError:
		status.Status = "error"
	}

	s.receiveEnvelope(e.ID)

	envelope := s.envelopes[e.ID]

	for _, ev := range envelope.Events {
		evS := storage.EventStatus{
			ID:             ev.ID,
			Type:           ev.Type,
			AvailableItems: len(ev.Items),
			TotalItems:     ev.ItemCount,
		}
		if evS.TotalItems != 0 && evS.TotalItems == evS.AvailableItems {
			evS.Status = "complete"
		} else {
			evS.Status = "receiving"
		}
		status.Events = append(status.Events, evS)
	}

	return status, nil
}

// ReadEnvelope reads as much as it can of an envelope, starting at 'position'
// The returned envelope should be suitable for encrypted nats transfert, ie
// its total jsonified size should be under maxsize
func (s *EnvelopeStorage) ReadEnvelope(id uuid.UUID, position storage.EnvelopeReadPos, maxsize int) (api.Envelope, storage.EnvelopeReadPos, error) {
	s.m.Lock()
	defer s.m.Unlock()

	envelope, ok := s.envelopes[id]
	if !ok {
		return api.Envelope{}, position, fmt.Errorf("No such envelope: %s", id)
	}

	if position.Start {
		for _, e := range envelope.Events {
			position.Events[e.ID] = storage.EventReadPos{}
		}
		position.Start = false
	}
	fragment := api.Envelope{
		ID: envelope.ID,
	}
	for _, event := range envelope.Events {
		fragment.EventIDs = append(fragment.EventIDs, event.ID)
	}

	// TODO in case we already know the envelope is small enough, return
	// it complete at once

	// approx size of the jsonified envelope before any event is added
	var estimateSize = 36 + len(fragment.EventIDs)*3

	position = position.Clone()

	tgtEvents := map[uuid.UUID]api.Event{}

	for _, ev := range envelope.Events {
		eventPos := position.Events[ev.ID]
		if eventPos.ItemCount < len(ev.Items) {
			eventPos.Index++
			tgtEvents[ev.ID] = api.Event{
				ID:        ev.ID,
				Type:      ev.Type,
				Index:     eventPos.Index,
				ItemCount: ev.ItemCount,
				Checksum:  ev.Checksum,
			}
			estimateSize += 36 + len(ev.Type) + 8
		}
		position.Events[ev.ID] = eventPos
	}

	for maxsize == 0 || estimateSize < maxsize {
		before := estimateSize
		for _, ev := range envelope.Events {
			pos := position.Events[ev.ID]
			if pos.ItemCount == len(ev.Items) {
				continue
			}
			tgt := tgtEvents[ev.ID]
			item := ev.Items[pos.ItemCount]
			tgt.Items = append(tgt.Items, item)
			pos.ItemCount++

			estimateSize += int(float64(len(item)) * 1.1)

			position.Events[ev.ID] = pos
			tgtEvents[ev.ID] = tgt
			if maxsize != 0 && estimateSize >= maxsize {
				break
			}
		}
		if estimateSize == before {
			break
		}
	}

	for _, ev := range tgtEvents {
		fragment.Events = append(fragment.Events, ev)
	}

	complete := true
	for _, ev := range envelope.Events {
		eventPos := position.Events[ev.ID]
		if ev.ItemCount == 0 || eventPos.ItemCount != ev.ItemCount {
			complete = false
		}
	}

	if complete {
		fragment.Last = true
		position.Complete = true
	}

	return fragment, position, nil
}

// Purge removes the envelopes fitered by the given filter
func (s *EnvelopeStorage) Purge(filter storage.EnvelopeFilter) {
	var todelete []uuid.UUID

	for id := range s.envelopes {
		todelete = append(todelete, filter([]uuid.UUID{id})...)
	}

	for _, id := range todelete {
		delete(s.envelopes, id)
		delete(s.receivers, id)
	}
}
