// Copyright 2016 Orus.io. All rights reserved.

package api

import (
	"github.com/m4rw3r/uuid"
)

const (
	// Msgbox is were envelopes are sent
	Msgbox = XbusAPIPrefix + "msgbox"

	// DefaultInput is the name of the default input
	DefaultInput = "default"
	// DefaultOutput is the name of the default output
	DefaultOutput = "default"
)

// EmitterMsgboxSubject returns the msgbox for an emitter
func EmitterMsgboxSubject(id uuid.UUID) string {
	return Msgbox + "." + id.String()
}

// ActorOutputSubject returns the subject for an actor output
func ActorOutputSubject(id uuid.UUID, output string) string {
	return Msgbox + "." + id.String() + ".output." + output
}

// ActorInputSubject returns the subject for an actor input
func ActorInputSubject(id uuid.UUID, output string) string {
	return Msgbox + "." + id.String() + ".input." + output
}

var (
	// EnvelopeMaxSize is an approximate maximum size of an envelope
	// A single Envelope should never exceed this size, or it could be
	// rejected by nats
	EnvelopeMaxSize = 750 * 1024 // 750K encrypted should be smaller than 1M
)

// Item is atomic data transported by the event
type Item []byte

// Event is the base unit manipulated by the graph
// The event structure transport a part of an Event,
// along with its checksum if it is the last part of
// the event
type Event struct {
	ID        uuid.UUID `json:"id"`
	Type      string    `json:"type,omitempty"`
	Index     int       `json:"index,omitempty"`
	Items     []Item    `json:"items,omitempty"`
	ItemCount int       `json:"count,omitempty"`
	Checksum  Checksum  `json:"checksum,omitempty"`
}

// ReceptionStatus is the reception status of an envelope
//go:generate enum-marshaler -type ReceptionStatus -prefix Reception -lower -tests
type ReceptionStatus int

const (
	// ReceptionReceiving is when the envelope is still incomplete and the
	// receiver waiting for the remains
	ReceptionReceiving ReceptionStatus = iota

	// ReceptionAccepted is when the envelope is completely received and
	// is guaranted to be processed.
	ReceptionAccepted

	// ReceptionError is when an error occured during reception of immediate
	// processing of the envelope
	ReceptionError
)

// EnvelopeContext contains informations used by the server to track the
// graph execution. It should be considered a blackbox from the worker point
// of view, which only copies it from input envelopes to output envelopes.
type EnvelopeContext struct {
	// ProcessID of the process the envelope is attached to.
	ProcessID uuid.UUID `json:"process_id"`

	// The graph node corresponding to the actor the envelope is being
	// sent to or received from.
	NodeID string `json:"node_id"`
}

// Envelope message contains a part (or all) of an envelope
type Envelope struct {
	ID uuid.UUID `json:"id"`

	// If provided, the complete list of event ids of the
	// envelope. Will be used to make sure all the pieces
	// were properly received
	EventIDs []uuid.UUID `json:"event_ids,omitempty"`

	// Any number of events from this envelope
	Events []Event `json:"events"`

	// true if this is the last envelope of a streamed envelope or
	// a unique envelope
	Last bool `json:"last"`

	// An actor must copy this value from input envelopes to
	// output envelopes
	// Emitters must not set the value
	Context EnvelopeContext `json:"context,omitempty"`
}

// EnvelopeAck is returned by actors after receiving an envelope
type EnvelopeAck struct {
	Response
	ID              uuid.UUID       `json:"id"`
	ReceptionStatus ReceptionStatus `json:"reception_status"`
	Reason          string          `json:"reason,omitempty"`
}

// NewEnvelopeAckError prepare a EnvelopeAck from an error
func NewEnvelopeAckError(id uuid.UUID, err error) EnvelopeAck {
	return EnvelopeAck{
		Response:        Response{Status: StatusClientError},
		ID:              id,
		ReceptionStatus: ReceptionError,
		Reason:          err.Error(),
	}
}

// CloseOutput returns a special envelope that closes an output
// without sending any data on it
func CloseOutput(context EnvelopeContext) Envelope {
	return Envelope{
		Context: context,
		Last:    true,
	}
}

// ForwardEnvelope builds a special envelope that indicates to
// the broker that this envelope is the same that was received
// on an input.
// Such an envelope can be sent only once per output, but can
// be sent before the complete envelope was received.
func ForwardEnvelope(envelope *Envelope) Envelope {
	return Envelope{
		ID:      envelope.ID,
		Context: envelope.Context,
		Last:    true,
	}
}
