// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"os"

	"github.com/spf13/cobra"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// registerCmd represents the "register" command.
var registerCmd = &cobra.Command{
	Use:   "register",
	Short: "Generate a CSR and ask Xbus for a registration.",
	Long: `Generate a Certificate Signing Request then send it over to Xbus through the
unencrypted "registration" channel. This xbusctl will only be allowed to send
Xbus actual commands once accepted on the Xbus side.
This command also outputs the command one has to run on the Xbus side to accept
the registration request.`,
	Run: register,
}

// register parses arguments to proceed with the registration.
func register(cmd *cobra.Command, args []string) {
	os.Exit(doRegister())
}

// doRegister generates a CSR then asks Xbus for a registration.
func doRegister() int {
	xbusClient, err := GetClient()
	if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}
	if err := xbusClient.Connect(); err != nil {
		Logger.Errorf("Cannot connect to server: %s", err)
		return 1
	}

	if err := xbusClient.CreateKey(false); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot create a private key %s", err)
			return 1
		}
	}

	// TODO find the actual hostname or get one in the configuration
	// TODO Read org and cn from the configuration
	if err := xbusClient.CreateCSR("localhost", "xbus", "xbusctl", false); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot generate a CSR: %s", err)
			return 1
		}
	}

	// Send the registration request.
	regStatus, err := xbusClient.Register()
	if err != nil {
		Logger.Errorf("Cannot send the registration request: %s", err)
		return 1
	}

	Logger.Noticef("Successfully sent the registration request.")

	// Say something depending on the registration status we have received in reply.
	switch regStatus {
	case api.RegistrationPending:
		Logger.Noticef(`The registration is pending approval. To approve from Xbus, run:

	xbusctl account accept %q
	xbusctl actor accept --id=%s

Then run this command again to save certificates:

	%s register`,
			// TODO tools.GetCSRHash(string(csrData), false), actorID,
            "xxx", "yyy",
            RootCmd.Use,
		)
	case api.RegistrationAccepted:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically accepted.",
		)
	case api.RegistrationRejected:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically rejected.",
		)
	}

	return 0
}

func init() {
	RootCmd.AddCommand(registerCmd)
}
