package demo

import (
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// NewRelayWorker is a RelayWorker factory
func NewRelayWorker(actor *xbus.Actor) xbus.ActorService {
	return &RelayWorker{
		actor,
		time.Duration(actor.GetIntSettingD("processing-duration", 1)) * time.Second,
		nil,
		nil,
	}
}

// RelayWorker is a worker that forward incoming envelopes to the default output
// The service parameters are:
// - processing-duration: time to wait between receive and send, in seconds
type RelayWorker struct {
	actor    *xbus.Actor
	interval time.Duration
	sub      *xbus.Subscription
	timers   []*time.Timer
}

// Startup initializes the subscription
func (e *RelayWorker) Startup() error {
	sub, err := e.actor.Subscribe(func(er *xbus.EnvelopeReceiver) {
		env, err := er.Complete(time.Second)
		if err != nil {
			e.actor.Log.Errorf("Reception error: %s", err)
			return
		}
		var timer *time.Timer
		timer = time.AfterFunc(e.interval, func() {
			status, err := e.actor.SendEnvelope("default", api.ForwardEnvelope(env))
			if status == api.ReceptionAccepted && err == nil {
				e.actor.Log.Noticef("Successfully forwarded the envelope")
			} else {
				e.actor.Log.Errorf("Something wrong happened. Status=%s, err=%s", status, err)
			}
			for i, t := range e.timers {
				if t == timer {
					e.timers = append(e.timers[:i], e.timers[i+1:]...)
				}
			}
		})
		e.timers = append(e.timers, timer)
	})
	if err != nil {
		return err
	}
	e.sub = sub
	return nil
}

// Shutdown delete the subscription
func (e *RelayWorker) Shutdown() error {
	for _, t := range e.timers {
		t.Stop()
	}
	err := e.sub.Unsubscribe()
	e.sub = nil
	return err
}

func init() {
	xbus.RegisterActorService("demo.relay", NewRelayWorker)
}
