package tools

import (
	"reflect"
	"time"

	"github.com/nats-io/go-nats"

	"github.com/nats-io/nuid"
)

// Conn A xbus wrapper for nats.EncodedConn. It add request versions that
// uses custom inboxes
type Conn struct {
	*nats.EncodedConn

	Namespace string // If the connection is a xbus client, its dedicated sub inbox
}

// RawRequestWithInbox is a nats.Conn.Request that takes the reply instead of
// calculating it automatically
func (c Conn) RawRequestWithInbox(subject, inbox string, data []byte, timeout time.Duration) (*nats.Msg, error) {
	ch := make(chan *nats.Msg, nats.RequestChanLen)

	s, err := c.Conn.ChanSubscribe(inbox, ch)
	if err != nil {
		return nil, err
	}
	s.AutoUnsubscribe(1)
	defer s.Unsubscribe()

	err = c.Conn.PublishRequest(subject, inbox, data)
	if err != nil {
		return nil, err
	}
	return s.NextMsg(timeout)
}

var emptyMsgType = reflect.TypeOf(&nats.Msg{})

// RequestWithInbox is a nats.EncodedConn.Request version that takes the reply
func (c Conn) RequestWithInbox(inbox, subject string, v, vPtr interface{}, timeout time.Duration) error {
	b, err := c.Enc.Encode(subject, v)
	if err != nil {
		return err
	}
	m, err := c.RawRequestWithInbox(subject, inbox, b, timeout)
	if err != nil {
		return err
	}
	if reflect.TypeOf(vPtr) == emptyMsgType {
		mPtr := vPtr.(*nats.Msg)
		*mPtr = *m
	} else {
		err = c.Enc.Decode(m.Subject, m.Data, vPtr)
	}
	return err

}

// RequestFor send a request for a node that has restricted access.
func (c Conn) RequestFor(dest, subject string, v, vPtr interface{}, timeout time.Duration) error {
	inbox := nats.InboxPrefix + dest + ".out." + nuid.Next()
	return c.RequestWithInbox(inbox, subject, v, vPtr, timeout)
}

// RequestFrom send a request from a node that has restrited access.
func (c Conn) RequestFrom(sender, subject string, v, vPtr interface{}, timeout time.Duration) error {
	inbox := nats.InboxPrefix + sender + ".in." + nuid.Next()
	return c.RequestWithInbox(inbox, subject, v, vPtr, timeout)
}

// Request defaults to RequestFrom if c.AccountID is defined, and to a classic inbox if not
func (c Conn) Request(subject string, v, vPtr interface{}, timeout time.Duration) error {
	if c.Namespace == "" {
		return c.EncodedConn.Request(subject, v, vPtr, timeout)
	}
	return c.RequestFrom(c.Namespace, subject, v, vPtr, timeout)
}
